from typing import TypeVar, Iterable, Iterator, Callable, Optional

###############################################################
# Iterator that allow the use of transformers using methods   #
# rather than functions                                       #
###############################################################

T = TypeVar("T")
R = TypeVar("R")


class ChainableIterator(Iterator[T]):
    def __init__(self, base: Iterator[T]):
        self.base = base

    def first(self) -> T:
        return next(self.base)

    def first_or_default(self, default: Optional[T] = None) -> Optional[T]:
        try:
            return next(self.base)
        except StopIteration:
            return default

    def filter(self, function: Callable[[T], bool]) -> "ChainableIterator[T]":
        iterator = filter(function, self.base)
        return ChainableIterator(iterator)

    def map(self, function: Callable[[T], R]) -> "ChainableIterator[R]":
        iterator = map(function, self.base)
        return ChainableIterator(iterator)

    def to_list(self):
        return list(self.base)

    # Warning: This consumes the iterator
    def count(self):
        return sum((1 for _ in self.base))

    def __next__(self) -> T:
        return next(self.base)


def citer(iterable: Iterable[T]) -> ChainableIterator[T]:
    return ChainableIterator(iter(iterable))




def hex_to_rgb(value):
    value = value.lstrip('#')
    val_len = len(value)
    return tuple(int(value[i:i + val_len // 3], 16) for i in range(0, val_len, val_len // 3))


def rgb_to_hex(rgb):
    return '#%02x%02x%02x' % rgb


def is_dark(color):
    if isinstance(color, str):
        color = hex_to_rgb(color)

    rgb = (color[0] / 255, color[1] / 255, color[2] / 255)
    lum = [rgb[i] for i in range(3)]

    for i in range(3):
        if lum[i] < 0.03928:
            lum[i] /= 12.92
        else: lum[i] = pow((lum[i] + 0.055) / 1.055, 2.4)

    return (0.2126 * lum[0] + 0.7152 * lum[1] + 0.0722 * lum[2]) <= 0.179



from typing import Tuple


class TextFormatting:
    ESCAPE = "\x1b"

    @staticmethod
    def combine(*formatting: "TextFormatting") -> "TextFormatting":
        codes = []
        for fmt in formatting:
            codes += list(fmt.get_codes())
        return TextFormatting(*codes)

    def __init__(self, *codes: int):
        self.codes = codes

    def get_codes(self) -> Tuple[int]:
        return self.codes

    def __str__(self) -> str:
        return f"{TextFormatting.ESCAPE}[" + ";".join([str(code) for code in self.get_codes()]) + "m"



import os
from typing import List


def is_null_or_empty(value) -> bool:
    return value is None or str(value) == ""


def is_null_or_white_space(value) -> bool:
    if value is None:
        return True
    return is_null_or_empty(str(value).strip())


def get_oneline_string(string: str) -> str:
    return string.replace("\n", "").replace("\r", "").replace("\t", "")


def list_to_string(objects: List) -> str:
    return ", ".join(map(str, objects))


def get_env(name: str, default = None) -> str:
    value = os.getenv(name, default)
    if value is None:
        raise EnvironmentError(f"The environment variable \"{name}\" is not defined")
    return value


def delete_file(path: str):
    if os.path.isfile(path):
        os.remove(path)



from typing import List, Tuple

from pursuitlib.console.textcomponent import TextComponent


class Table:
    COLUMN_PADDING = 1

    def __init__(self):
        self.rows: List[Tuple[any]] = []

    def append_row(self, *columns: any):
        self.rows.append(columns)

    def get_row_count(self) -> int:
        return len(self.rows)

    def get_column_count(self) -> int:
        if len(self.rows) == 0:
            return 0
        return max([len(row) for row in self.rows])

    def get_column_widths(self) -> List[int]:
        widths = [0] * self.get_column_count()
        for row in self.rows:
            for col, component in enumerate(row):
                widths[col] = max(widths[col], len(component))
        return widths

    def as_lines(self) -> List[TextComponent]:
        lines = []
        widths = self.get_column_widths()
        for row in self.rows:
            line = TextComponent(None)
            for col, component in enumerate(row):
                padding = widths[col] - len(component)
                if col < len(widths) - 1:  # Do not add column padding after the last column
                    padding += Table.COLUMN_PADDING
                line.append(component, ' ' * padding)
            lines.append(line)
        return lines

    def as_text(self) -> TextComponent:
        text = TextComponent()
        lines = self.as_lines()
        for i, line in enumerate(lines):
            if i < len(lines) - 1:
                text.appendline(line)
            else: text.append(line)  # Do not append a newline at the end of the table
        return text

    def __str__(self) -> str:
        return self.as_text().render()



from typing import TypeVar

T = TypeVar("T")


# This decorator can be used to easily create decorators that
# can optionally take arguments
def decorator(function):
    def wrapper(*args, **kwargs):
        # If the decorator is used without parenthesis, the first argument
        # is the function that needs to be passed to "decorator"
        if len(args) == 1 and callable(args[0]) and len(kwargs) == 0:
            fn = args[0]
            return function(fn)
        else: return lambda f: function(f, *args, **kwargs)
    return wrapper


# This decorator can be used to cache the result of a method
def cached(function: T) -> T:
    def wrapper(self, *args, force_update: bool = False, **kwargs):
        cached_field = f"__cached_{function.__name__}"

        # Use the cached value when available
        if not force_update and hasattr(self, cached_field):
            return getattr(self, cached_field)

        result = function(self, *args, **kwargs)
        setattr(self, cached_field, result)  # Update the cache
        return result
    return wrapper



from typing import Optional, Union

from pursuitlib.console.textcolor import TextColor
from pursuitlib.console.textstyle import TextStyle


class TextComponent:
    @staticmethod
    def concat(*components) -> "TextComponent":
        return TextComponent(None, *components)

    def __init__(self, style: Optional[Union[TextStyle, TextColor]] = None, *children):
        self.parent: Optional[TextComponent] = None

        if isinstance(style, TextStyle):
            self.style = style
        elif isinstance(style, TextColor):
            self.style = TextStyle(color=style)
        else: self.style = TextStyle()

        self.children = list(children)

    def set_parent(self, parent: "TextComponent"):
        self.parent = parent

    def get_parent(self) -> Optional["TextComponent"]:
        return self.parent

    def set_style(self, style: TextStyle):
        self.style = style

    def get_style(self) -> TextStyle:
        return self.style

    def append(self, *children):
        for child in children:
            if isinstance(child, TextComponent):
                child.set_parent(self)
            self.children.append(child)

    def appendline(self, *children):
        self.append(*children, '\n')

    def length(self) -> int:
        length = 0
        for child in self.children:
            if isinstance(child, TextComponent):
                length += child.length()
            else: length += len(str(child))
        return length

    def render(self) -> str:
        result = f"{self.get_style().as_formatting(self)}"
        for child in self.children:
            if isinstance(child, TextComponent):
                result += child.render()
                # Reset the formatting to this component's formatting
                result += f"{self.get_style().as_formatting(self)}"
            else: result += str(child)

        # Always finish by resetting the text formatting
        if self.parent is None:
            result += str(TextStyle.FORMAT_RESET)
        return result

    def __len__(self) -> int:
        return self.length()

    def __str__(self) -> str:
        return self.render()



import sys
from typing import Optional, Union, IO

from pursuitlib.console.textcolor import TextColor
from pursuitlib.console.textcolors import TextColors
from pursuitlib.console.textcomponent import TextComponent
from pursuitlib.console.textstyle import TextStyle

##################
# Console writer #
##################


def write(item, style: Optional[Union[TextStyle, TextColor]] = None):
    write_to(sys.stdout, item, style)


def write_line(item="", style: Optional[Union[TextStyle, TextColor]] = None):
    write(f"{item}\n", style)


def write_err(item, style: Optional[Union[TextStyle, TextColor]] = None):
    if style is None:
        style = TextColors.LIGHT_RED
    write_to(sys.stderr, item, style)


def write_err_line(item="", style: Optional[Union[TextStyle, TextColor]] = None):
    write_err(f"{item}\n", style)


def write_to(stream: IO, item, style: Optional[Union[TextStyle, TextColor]] = None):
    if style is None:
        stream.write(str(item))
    else:
        if isinstance(style, TextColor):
            style = TextStyle(color=style)
        stream.write(str(TextComponent(style, item)))


def confirm(prompt: str, default_answer: Optional[bool] = None) -> bool:
    while True:
        if default_answer is None:
            choices = "[y/n]"
        elif default_answer:
            choices = "[Y/n]"
        else:  # not default_answer
            choices = "[y/N]"

        result = input(f"{prompt} {choices} ").strip().lower()

        if result == "" and default_answer is not None:
            return default_answer
        elif result == "y":
            return True
        elif result == "n":
            return False



from pursuitlib.console.textformatting import TextFormatting


class TextColor(TextFormatting):
    @staticmethod
    def standard(code: int) -> "TextColor":
        return TextColor(code)

    @staticmethod
    def rgb(color: int) -> "TextColor":
        r = (color >> 16) & 0xFF
        g = (color >> 8) & 0xFF
        b = color & 0xFF
        return TextColor(38, 2, r, g, b)



from typing import Optional

from decimal import Decimal


def parseint(value, default: Optional[int] = None) -> Optional[int]:
    if value is not None:
        try:
            return int(value)
        except ValueError:
            pass

    return default


def parsefloat(value, default: Optional[float] = None) -> Optional[float]:
    if value is not None:
        try:
            return float(value)
        except ValueError:
            pass

    return default


def parseas(type_name: str, value, default=None):
    if value is not None:
        try:
            if type_name == "str":
                return str(value)
            elif type_name == "int":
                return int(value)
            elif type_name == "float":
                return float(value)
            elif type_name == "Decimal":
                return Decimal(value)
        except ValueError:
            pass

    return default



from enum import Enum


class IndexedEnum(Enum):
    @classmethod
    def values(cls):
        return list(cls)

    def __init__(self, *args):
        self.ordinal = len(type(self))

    def __str__(self) -> str:
        return self.name

    def __int__(self) -> int:
        return self.ordinal



from typing import Callable


def not_implemented(obj: object, method: Callable) -> NotImplementedError:
    return NotImplementedError(f"Method '{method.__name__}' is not implemented for '{type(obj).__name__}'")



from setuptools import setup, find_packages
from pathlib import Path

directory = Path(__file__).parent

setup(
    name="pursuitlib",
    version="0.4.0",
    packages=find_packages(),
    entry_points={},
    author="Pursuit",
    author_email="fr.pursuit@gmail.com",
    description="Provides utility functions",
    long_description=(directory / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://gitlab.com/frPursuit/pursuitlib-python",
    license="All rights reserved",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8"
)



from typing import Optional, TYPE_CHECKING

from pursuitlib.console.textcolor import TextColor
from pursuitlib.console.textformatting import TextFormatting

if TYPE_CHECKING:
    from pursuitlib.console.textcomponent import TextComponent


class TextStyle:
    FORMAT_RESET = TextFormatting(0)
    FORMAT_BOLD = TextFormatting(1)
    FORMAT_UNDERLINE = TextFormatting(4)
    FORMAT_STRIKETHROUGH = TextFormatting(9)

    def __init__(self, color: Optional[TextColor] = None, bold: Optional[bool] = None, underline: Optional[bool] = None, strikethrough: Optional[bool] = None):
        self.color = color
        self.bold = bold
        self.underline = underline
        self.strikethrough = strikethrough

    def get_color(self) -> Optional[TextColor]:
        return self.color

    def is_bold(self) -> Optional[bool]:
        return self.bold

    def is_underline(self) -> Optional[bool]:
        return self.underline

    def is_strikethrough(self) -> Optional[bool]:
        return self.strikethrough

    def get_effective_color(self, component: Optional["TextComponent"] = None) -> Optional[TextColor]:
        if self.color is not None:
            return self.color
        if component is not None and component.parent is not None:
            return component.parent.get_style().get_effective_color(component.parent)
        return None  # Use the terminal's default color

    def is_effectively_bold(self, component: Optional["TextComponent"] = None) -> bool:
        if self.bold is not None:
            return self.bold
        if component is not None and component.parent is not None:
            return component.parent.get_style().is_effectively_bold(component.parent)
        return False

    def is_effectively_underline(self, component: Optional["TextComponent"] = None) -> bool:
        if self.underline is not None:
            return self.underline
        if component is not None and component.parent is not None:
            return component.parent.get_style().is_effectively_underline(component.parent)
        return False

    def is_effectively_strikethrough(self, component: Optional["TextComponent"] = None) -> bool:
        if self.strikethrough is not None:
            return self.strikethrough
        if component is not None and component.parent is not None:
            return component.parent.get_style().is_effectively_strikethrough(component.parent)
        return False

    def as_formatting(self, component: Optional["TextComponent"] = None) -> TextFormatting:
        color = self.get_effective_color(component)
        bold = self.is_effectively_bold(component)
        underline = self.is_effectively_underline(component)
        strikethrough = self.is_effectively_strikethrough(component)

        # We want the TextFormatting element to be independant from the previous text formatting,
        # so we start by resetting it
        formatting = [TextStyle.FORMAT_RESET]
        if color is not None:
            formatting.append(color)
        if bold:
            formatting.append(TextStyle.FORMAT_BOLD)
        if underline:
            formatting.append(TextStyle.FORMAT_UNDERLINE)
        if strikethrough:
            formatting.append(TextStyle.FORMAT_STRIKETHROUGH)
        return TextFormatting.combine(*formatting)



from pursuitlib.console.textcolor import TextColor


class TextColors:
    BLACK = TextColor.standard(30)
    DARK_RED = TextColor.standard(31)
    DARK_GREEN = TextColor.standard(32)
    DARK_YELLOW = TextColor.standard(33)
    DARK_BLUE = TextColor.standard(34)
    DARK_MAGENTA = TextColor.standard(35)
    DARK_CYAN = TextColor.standard(36)
    LIGHT_GRAY = TextColor.standard(37)
    DARK_GRAY = TextColor.standard(90)
    LIGHT_RED = TextColor.standard(91)
    LIGHT_GREEN = TextColor.standard(92)
    LIGHT_YELLOW = TextColor.standard(93)
    LIGHT_BLUE = TextColor.standard(94)
    LIGHT_MAGENTA = TextColor.standard(95)
    LIGHT_CYAN = TextColor.standard(96)
    WHITE = TextColor.standard(97)
